from ._cli import cli
from ._parser import FASTAItem, FASTAParser, ParsingError, open_fasta
from ._testit import test
from ._writer import FASTAWriter
from ._version import __version__


__all__ = [
    "FASTAItem",
    "FASTAParser",
    "FASTAWriter",
    "ParsingError",
    "__version__",
    "cli",
    "open_fasta",
    "test",
]



import pytest
from fasta_reader import open_fasta, ParsingError


def _test_fasta_correct(filepath):
    pass

    deflines = ["ID1", "ID2", "ID3", "ID4"]
    sequences = ["GAGUUA", "CAUAACAAATT", "AAGAA", "AAGAA"]

    f = open_fasta(filepath)
    item = f.read_item()
    assert item.defline == deflines[0]
    assert item.sequence == sequences[0]

    item = f.read_item()
    assert item.defline == deflines[1]
    assert item.sequence == sequences[1]

    item = f.read_item()
    assert item.defline == deflines[2]
    assert item.sequence == sequences[2]

    item = f.read_item()
    assert item.defline == deflines[3]
    assert item.sequence == sequences[3]

    with pytest.raises(StopIteration):
        f.read_item()

    f.close()

    f = open_fasta(filepath)
    for i, item in enumerate(f):
        assert item.defline == deflines[i]
        assert item.sequence == sequences[i]
    f.close()

    f = open_fasta(filepath)
    items = f.read_items()
    for i in range(len(deflines)):
        assert items[i].defline == deflines[i]
        assert items[i].sequence == sequences[i]
    f.close()

    with open_fasta(filepath) as f:
        for i, item in enumerate(f):
            assert item.defline == deflines[i]
            assert item.sequence == sequences[i]

    with open_fasta(filepath) as f:
        f.close()


def test_fasta_correct1(correct1):
    _test_fasta_correct(correct1)


def test_fasta_correct2(correct2):
    _test_fasta_correct(correct2)


def test_fasta_damaged(damaged1, damaged2, damaged3):

    with open_fasta(damaged1) as f:
        with pytest.raises(ParsingError):
            f.read_item()

    with open_fasta(damaged2) as f:
        with pytest.raises(ParsingError):
            f.read_item()

    with open_fasta(damaged3) as f:
        f.read_item()
        with pytest.raises(ParsingError):
            f.read_item()



import pathlib
from typing import IO, Iterator, List, NamedTuple, Union

from more_itertools import peekable


class ParsingError(Exception):
    pass


FASTAItem = NamedTuple("FASTAItem", [("defline", str), ("sequence", str)])


class FASTAParser:
    def __init__(self, file: Union[str, pathlib.Path, IO[str]]):
        if isinstance(file, str):
            file = pathlib.Path(file)

        if isinstance(file, pathlib.Path):
            file = open(file, "r")

        self._file = file
        self._lines = peekable(line for line in file)

    def read_item(self) -> FASTAItem:
        """
        Get the next item.
        """
        defline = self._next_defline()
        sequence = self._next_sequence()
        return FASTAItem(defline, sequence)

    def read_items(self) -> List[FASTAItem]:
        """
        Get the list of all items.
        """
        return list(self)

    def close(self):
        """
        Close the associated stream.
        """
        self._file.close()

    def _next_defline(self) -> str:
        while True:
            line = next(self._lines)
            if line == "":
                raise StopIteration

            line = line.strip()
            if line.startswith(">"):
                return line[1:]
            if line != "":
                raise ParsingError()

    def _next_sequence(self) -> str:
        lines = []
        while True:
            line = next(self._lines)
            if line == "":
                raise ParsingError()

            line = line.strip()
            if not line.startswith(">"):
                lines.append(line)
                if self._sequence_continues():
                    continue
                return "".join(lines)
            if line != "":
                raise ParsingError()

    def _sequence_continues(self):
        try:
            next_line = self._lines.peek()
        except StopIteration:
            return False

        if next_line == "":
            return False
        next_line = next_line.strip()
        return len(next_line) > 0 and not next_line.startswith(">")

    def __iter__(self) -> Iterator[FASTAItem]:
        while True:
            try:
                yield self.read_item()
            except StopIteration:
                return

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        del exception_type
        del exception_value
        del traceback
        self.close()


def open_fasta(file: Union[str, pathlib.Path, IO[str]]) -> FASTAParser:
    """
    Open a FASTA file.

    Parameters
    ----------
    file : Union[str, pathlib.Path, IO[str]]
        File path or IO stream.

    Returns
    -------
    parser : FASTAParser
        FASTA parser.
    """
    return FASTAParser(file)



import pytest


@pytest.fixture
def correct1(tmp_path):
    return _write_file(tmp_path, "correct1.seq")


@pytest.fixture
def correct2(tmp_path):
    return _write_file(tmp_path, "correct2.seq")


@pytest.fixture
def damaged1(tmp_path):
    return _write_file(tmp_path, "damaged1.seq")


@pytest.fixture
def damaged2(tmp_path):
    return _write_file(tmp_path, "damaged2.seq")


@pytest.fixture
def damaged3(tmp_path):
    return _write_file(tmp_path, "damaged3.seq")


def _write_file(path, filename):
    import importlib_resources as pkg_resources
    import fasta_reader

    text = pkg_resources.read_text(fasta_reader.test, filename)

    with open(path / filename, "w") as f:
        f.write(text)

    return path / filename



from statistics import mean

import click

from ._parser import open_fasta
from ._version import __version__


@click.command(context_settings=dict(help_option_names=["-h", "--help"]))
@click.version_option(__version__)
@click.argument("fasta", type=click.File("r"))
def cli(fasta):
    """
    Show information about FASTA files.
    """
    with open_fasta(fasta) as parser:
        items = parser.read_items()

    nitems = len(items)
    abc = set([])
    seq_lens = []
    for item in items:
        seq_lens.append(len(item.sequence))
        abc |= set(item.sequence)

    usymbols = "".join(sorted(list(abc)))
    click.echo(f"Number of sequences: {nitems}")
    click.echo(f"Unique sequence symbols: {usymbols}")
    click.echo(f"Average sequence length: {mean(seq_lens)}")



import pathlib
from typing import IO, Union


class FASTAWriter:
    def __init__(self, file: Union[str, pathlib.Path, IO[str]]):
        if isinstance(file, str):
            file = pathlib.Path(file)

        if isinstance(file, pathlib.Path):
            file = open(file, "w")

        self._file = file

    def write_item(self, defline: str, sequence: str):
        """
        Write item.
        """
        self._file.write(">" + defline + "\n")
        self._file.write(sequence + "\n")

    def close(self):
        """
        Close the associated stream.
        """
        self._file.close()

    def __enter__(self):
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        del exception_type
        del exception_value
        del traceback
        self.close()
